#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>
#include "widgets.h"

struct wgt_window {
	HWND handle;
	struct wgt_rect rect;

	HDC dc;
	char cname[32];
	struct wgt_widget *wlist;
};

struct wgt_widget {
	HWND handle;
	struct wgt_rect rect;

	struct wgt_window *win;
	struct wgt_widget *next;

	char **itemlist;
	int num_items;

	wgt_callback cb_click;
	wgt_callback cb_modify;
};


static LRESULT WINAPI event_handler(HWND win, unsigned int msg, WPARAM wparam, LPARAM lparam);
static struct wgt_widget *find_widget(struct wgt_window *win, HWND handle);

struct wgt_window *wgt_window(const char *title, int width, int height)
{
	struct wgt_window *win;
	int tlen;
	WNDCLASS wc;
	HINSTANCE hinst;
	DWORD err;
	RECT rect;

	if(!(win = malloc(sizeof *win))) {
		fprintf(stderr, "wgl_create_window: failed to allocate window structure\n");
		return 0;
	}
	win->wlist = 0;

	tlen = strlen(title);
	if(tlen > sizeof win->cname - 10) {
		tlen = sizeof win->cname - 10;
	}
	strcpy(win->cname, "WGTCLASS-");
	memcpy(win->cname + 9, title, tlen);
	win->cname[tlen + 9] = 0;

	hinst = GetModuleHandle(0);

	memset(&wc, 0, sizeof wc);
	wc.hInstance = hinst;
	wc.hbrBackground = GetSysColorBrush(COLOR_3DFACE);
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = LoadIcon(0, IDI_APPLICATION);
	wc.lpszClassName = win->cname;
	wc.lpfnWndProc = event_handler;
	wc.cbWndExtra = 4;
	if(!RegisterClass(&wc)) {
		fprintf(stderr, "wgt_create_window: failed to register window class\n");
		return 0;
	}

	if(!(win->handle = CreateWindow(win->cname, title, WS_OVERLAPPEDWINDOW,
			CW_USEDEFAULT, CW_USEDEFAULT, width, height, 0, 0, hinst, 0))) {
		fprintf(stderr, "wgt_create_window: failed to create window\n");
		UnregisterClass(title, hinst);
		return 0;
	}
	win->dc = GetDC(win->handle);
	SetLastError(0);
	if(!SetWindowLong(win->handle, GWL_USERDATA, (long)win) && (err = GetLastError())) {
		fprintf(stderr, "wgt_create_window: failed to store window user data (err: %d)\n", (int)err);
		UnregisterClass(title, hinst);
		DestroyWindow(win->handle);
		return 0;
	}
	ShowWindow(win->handle, 1);

	GetWindowRect(win->handle, &rect);

	win->rect.x = rect.left;
	win->rect.y = rect.top;
	win->rect.width = rect.right - rect.left;
	win->rect.height = rect.bottom - rect.top;

	return win;
}

void wgt_destroy_window(struct wgt_window *win)
{
	while(win->wlist) {
		struct wgt_widget *w = win->wlist;
		win->wlist = win->wlist->next;
		wgt_destroy_widget(w);
		free(w);
	}

	ReleaseDC(win->handle, win->dc);
	DestroyWindow(win->handle);
	UnregisterClass(win->cname, GetModuleHandle(0));

	free(win);
}

void wgt_destroy_widget(struct wgt_widget *w)
{
	DestroyWindow(w->handle);
	free(w->itemlist);
}

struct wgt_window *wgt_widget_window(struct wgt_widget *w)
{
	return w->win;
}

struct wgt_rect *wgt_widget_rect(struct wgt_widget *w, struct wgt_rect *rect)
{
	if(rect) {
		*rect = w->rect;
	}
	return &w->rect;
}

int wgt_xpos_after(struct wgt_widget *w, int pad)
{
	if(pad == WGT_AUTO) {
		pad = 8;
	}
	return w->rect.x + w->rect.width + pad;
}

int wgt_ypos_after(struct wgt_widget *w, int pad)
{
	if(pad == WGT_AUTO) {
		pad = 8;
	}
	return w->rect.y + w->rect.height + pad;
}

int wgt_string_size(struct wgt_window *win, const char *s, struct wgt_rect *rect)
{
	SIZE sz;

	GetTextExtentPoint32(win->dc, (char*)s, strlen(s), &sz);
	if(rect) {
		rect->x = rect->y = 0;
		rect->width = sz.cx;
		rect->height = sz.cy;
	}
	return sz.cx;
}


struct wgt_widget *wgt_label(struct wgt_window *win, const char *text, int x, int y)
{
	struct wgt_widget *w;
	struct wgt_rect textsz;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_label: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	wgt_string_size(win, text, &textsz);

	if(!(w->handle = CreateWindow("STATIC", text, WS_CHILD | WS_VISIBLE | SS_SIMPLE,
			x, y, textsz.width, textsz.height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_label: failed to create window\n");
		free(w);
		return 0;
	}

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = textsz.width;
	w->rect.height = textsz.height;

	w->next = win->wlist;
	win->wlist = w;
	return w;
}

struct wgt_widget *wgt_button(struct wgt_window *win, const char *text, int x, int y,
			int width, int height, wgt_callback clickfunc)
{
	struct wgt_widget *w;
	struct wgt_rect textsz;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_button: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	if(width < 0 || height < 0) {
		wgt_string_size(win, text, &textsz);
		if(width < 0) width = textsz.width * 5 / 3;
		if(height < 0) height = textsz.height * 10 / 7;
	}

	if(!(w->handle = CreateWindow("BUTTON", text, WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
			x, y, width, height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_button: failed to create window\n");
		free(w);
		return 0;
	}

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = width;
	w->rect.height = height;

	w->cb_click = clickfunc;
	w->next = win->wlist;
	win->wlist = w;
	return w;
}

struct wgt_widget *wgt_checkbox(struct wgt_window *win, const char *text, int on,
			int x, int y, int width, int height, wgt_callback modfunc)
{
	struct wgt_widget *w;
	struct wgt_rect textsz;
	int check_width;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_checkbox: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	check_width = GetSystemMetrics(SM_CXMENUCHECK);

	if(width < 0 || height < 0) {
		wgt_string_size(win, text, &textsz);
		if(width < 0) width = textsz.width + check_width + 10;
		if(height < 0) height = textsz.height;
	}

	if(!(w->handle = CreateWindow("BUTTON", text, WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX,
			x, y, width, height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_checkbox: failed to create window\n");
		free(w);
		return 0;
	}

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = width;
	w->rect.height = height;

	w->cb_click = modfunc;	/* BN_CLICKED is sent for checkbox state changes */
	w->next = win->wlist;
	win->wlist = w;
	return w;
}

struct wgt_widget *wgt_combo(struct wgt_window *win, const char **items, int num_items,
			int sel, int x, int y, int width, int height, wgt_callback modfunc)
{
	int i, max_width, max_height, sum_height, res;
	struct wgt_widget *w;
	struct wgt_rect textsz;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_combo: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	if(!(w->itemlist = malloc(num_items * sizeof *w->itemlist))) {
		fprintf(stderr, "wgt_combo: failed to allocate item list\n");
		free(w);
		return 0;
	}

	max_width = max_height = sum_height = 0;

	for(i=0; i<num_items; i++) {
		wgt_string_size(win, items[i], &textsz);
		if(textsz.width > max_width) max_width = textsz.width;
		if(textsz.height > max_height) max_height = textsz.height;
		sum_height += textsz.height;
	}
	if(width < 0) width = max_width * 3 / 2;
	if(height < 0) height = sum_height * 2;


	if(!(w->handle = CreateWindow("COMBOBOX", items[0],
			WS_CHILD | WS_VISIBLE | CBS_DROPDOWNLIST | CBS_HASSTRINGS,
			x, y, width, height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_combo: failed to create window\n");
		free(w);
		return 0;
	}

	for(i=0; i<num_items; i++) {
		if(!(w->itemlist[i] = strdup(items[i]))) {
			fprintf(stderr, "wgt_combo: failed to allocate item\n");
			while(--i >= 0) free(w->itemlist[i]);
			DestroyWindow(w->handle);
			free(w->itemlist);
			free(w);
			return 0;
		}
		if((res = SendMessage(w->handle, CB_ADDSTRING, 0, (long)items[i])) != i) {
			fprintf(stderr, "wgt_combo: failed to add item\n");
		}
	}
	w->num_items = num_items;

	if(sel < 0) sel = 0;
	if(sel >= num_items) sel = num_items - 1;
	SendMessage(w->handle, CB_SETCURSEL, sel, 0);

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = width;
	w->rect.height = max_height;

	w->cb_modify = modfunc;
	w->next = win->wlist;
	win->wlist = w;
	return w;
}

int wgt_combo_selected(struct wgt_widget *w)
{
	int sel = SendMessage(w->handle, CB_GETCURSEL, 0, 0);
	if(sel == CB_ERR) return -1;
	return sel;
}

const char *wgt_get_combo_item(struct wgt_widget *w, int idx)
{
	if(idx < 0 || idx >= w->num_items) {
		return 0;
	}
	return w->itemlist[idx];
}

static LRESULT WINAPI event_handler(HWND wnd, unsigned int msg, WPARAM wparam, LPARAM lparam)
{
	int ncode;
	struct wgt_widget *w;
	struct wgt_window *win;

	win = (struct wgt_window*)GetWindowLong(wnd, GWL_USERDATA);

	switch(msg) {
	case WM_CLOSE:
	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_COMMAND:
		if(!win || !(w = find_widget(win, (HWND)lparam))) {
			break;
		}
		ncode = HIWORD(wparam);

		switch(ncode) {
		case BN_CLICKED:
			if(w->cb_click) {
				w->cb_click(w);
			}
			break;

		case CBN_SELCHANGE:
			if(w->cb_modify) {
				w->cb_modify(w);
			}
			break;

		default:
			break;
		}

	default:
		return DefWindowProc(wnd, msg, wparam, lparam);
	}
	return 0;
}

static struct wgt_widget *find_widget(struct wgt_window *win, HWND handle)
{
	struct wgt_widget *w;

	w = win->wlist;
	while(w) {
		if(w->handle == handle) return w;
		w = w->next;
	}
	return 0;
}
